/**
 * Driver that handles the EXTSMI# interrupt on the xbox.
 * Makes it possible to use the eject-button without the xbox rebooting...
 *
 * smbus-command sequence to prevent reboot from cromwell.
 *
 * Changelog:
 *  2003-01-14 Anders Gustafsson <andersg@0x63.nu>
 *             initial version
 *
 * Todo: add errorhandling!
 *
 */




#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <asm/io.h>

#include <linux/i2c.h>



/* FIXME! -> .h */

#define PIC_ADDRESS 0x10

#define SMC_CMD_TRAY_STATE 0x03
#define SMC_CMD_EJECT 0x0C
#define SMC_CMD_INTERRUPT_REASON 0x11
#define SMC_CMD_RESET_ON_EJECT 0x19
#define SMC_CMD_LED_MODE 0x07
#define SMC_CMD_LED_REGISTER 0x08

#define SMC_SUBCMD_EJECT_EJECT 0x00
#define SMC_SUBCMD_EJECT_LOAD 0x01
#define SMC_SUBCMD_RESET_ON_EJECT_ENABLE 0x00
#define SMC_SUBCMD_RESET_ON_EJECT_DISABLE 0x01

/* interrupt causes */
#define POWERDOWN_MASK (1<<0)
#define TRAYCLOSED_MASK (1<<1)
#define TRAYOPENING_MASK (1<<2)
#define AVPLUGGED_MASK (1<<3)
#define AVUNPLUGGED_MASK (1<<4)
#define TRAYBUTTON_MASK (1<<5)
#define TRAYCLOSING_MASK (1<<6)
#define UNKNOWN_MASK (1<<7)

#define IRQ 12
#define DRIVER_NAME "extsmi"

#define DEBUG(a ...) printk(KERN_INFO a)


/* just some crap */
static char dev[]="test";

#define BASE 0x8000

/* Power Management 1 Enable Register */
#define PM02 (BASE+0x02)

/* Power Management 1 Control Register */
#define PM04 (BASE+0x04)

/* ACPI GP Status Register */
#define PM20 (BASE+0x20)

/* ACPI GP Enable Register */
#define PM22 (BASE+0x22)
# define EXTSMI_EN_MASK 0x0002

/* Global SMI Enable Register */
#define PM2A (BASE+0x2A)


static DECLARE_MUTEX(extsmi_sem);
static DECLARE_COMPLETION(extsmi_exited);
static int extsmi_pid=0;

static int extsmi_attach_adapter(struct i2c_adapter *adap);

static struct i2c_driver extsmi_driver = {
	name:		"i2c xbox-extsmi driver",
	id:		I2C_DRIVERID_I2CDEV,
	flags:		I2C_DF_DUMMY,
	attach_adapter:	extsmi_attach_adapter,
};

static struct i2c_adapter *xbox_adap;

static struct i2c_client extsmi_client = {
	name:		"I2C xbox-extsmi client",
	id:		1,
	flags:		0,
	addr:		PIC_ADDRESS,
	adapter:	NULL,
	driver:		&extsmi_driver,
};


static int extsmi_attach_adapter(struct i2c_adapter *adap)
{
	int i;

	if ((i = i2c_adapter_id(adap)) < 0) {
		printk("i2c-dev.o: Unknown adapter ?!?\n");
		return -ENODEV;
	}

	/* FIXME! XXX! This is bogus, will break with more than one adaptor */
	if (! xbox_adap) {
		xbox_adap = adap;
		extsmi_client.adapter=adap;
		printk(KERN_INFO DRIVER_NAME ": Using '%s'!\n",adap->name);
	} else {
		/* This is actually a detach_adapter call! */
		xbox_adap=NULL;
	}

	return 0;
}



static void extsmi_interupt(int i,void *dev,struct pt_regs *r){
	int reason;

	reason=inw(0x8020);
	outw(reason,0x8020); /* ack  IS THIS NEEDED? */
	if(reason&0x2){
		DEBUG("EXTSMI#\n");

		/* wake up thread */
		up(&extsmi_sem);
		DEBUG("Interrupt: 0x%02x\n",reason);
	}
/*
	reason=inw(0x8020);
	DEBUG("Interrupt %d 0x%04x\n",hits,reason);
*/

/*	 Stop irq-storm
	if(hits==400){
		free_irq(12,dev);
	}
*/
}



/**
 * Process a event. This is run i process-context.
 */
static void extsmi_process(void){
	int reason;
	
	if(extsmi_client.adapter==NULL){
		printk("bh with no client!?\n");
		return;
	}
	
	reason=i2c_smbus_read_byte_data(&extsmi_client,SMC_CMD_INTERRUPT_REASON);
	if(reason&TRAYBUTTON_MASK){ /* Tray button! Respond to prevent reboot! */
		i2c_smbus_write_byte_data(&extsmi_client,0x0d,0x04);
		i2c_smbus_write_byte_data(&extsmi_client,0x00,0x0c);
		DEBUG("Button<tm> pressed\n");

		/* eject? */
#ifndef NO_EJECT
		i2c_smbus_write_byte_data(&extsmi_client, SMC_CMD_RESET_ON_EJECT, SMC_SUBCMD_RESET_ON_EJECT_DISABLE);
		i2c_smbus_write_byte_data(&extsmi_client, SMC_CMD_EJECT, SMC_SUBCMD_EJECT_EJECT);
#endif
	}
	
	DEBUG("bh-reason: 0x%02x\n",reason);
}


static int extsmi_thread(void *data){

	daemonize();
	reparent_to_init();
	strcpy(current->comm, "xbox_extsmi");


	do {
		extsmi_process();
		down_interruptible(&extsmi_sem);
	} while (!signal_pending(current));
	
         complete_and_exit(&extsmi_exited, 0);

}


static int extsmi_init(void){
	int res;
	int pid;
	
        pid = kernel_thread(extsmi_thread, NULL,
			    CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	if (pid < 0) {
		return pid;
	}

	extsmi_pid = pid;

	if ((res = i2c_add_driver(&extsmi_driver))) {
		printk(KERN_ERR DRIVER_NAME ": Driver registration failed, module not inserted.\n");
		/* FIXME! CLEANUP!! */
		return res;
	}

	/* this shuts a lot of interrupts off! */
	outw(inw(0x80e2)&0xf8c7,0x80e2);
	outw(0,0x80ac);
	outb(0,0x8025);
	outw(EXTSMI_EN_MASK,PM22); /* enable the EXTSMI# interupt! */
	outw(0,PM02);
	outw(0,0x80d8);

	outb(1,PM04); /* enable sci interrupts! */

	/* FIXME! retval! */
	request_irq(IRQ,extsmi_interupt,SA_INTERRUPT|SA_SHIRQ /* |SA_SAMPLE_RANDOM */,
		    "xbox_extsmi",dev);

	
	/* LED blinking
	  i2c_smbus_write_byte_data(&i2cxbox_client,SMC_CMD_LED_MODE,1);
	  i2c_smbus_write_byte_data(&i2cxbox_client,SMC_CMD_LED_REGISTER,0xc6);
	*/

	return 0;
}

static void extsmi_exit(void){
	int res;
	free_irq(IRQ,dev);

	/* Kill the thread */
	res = kill_proc(extsmi_pid, SIGTERM, 1);
	wait_for_completion(&extsmi_exited);

	if ((res = i2c_del_driver(&extsmi_driver))) {
		printk(KERN_ERR DRIVER_NAME ": Driver deregistration failed, "
		       "module not removed.\n");
		return;
	}
	
	return;
}

module_init(extsmi_init);
module_exit(extsmi_exit);

MODULE_AUTHOR("Anders Gustafsson <andersg@0x63.nu>");
MODULE_LICENSE("GPL");
