/*
 * X-Box gamepad mousedriver - v0.0.2
 *
 * Copyright (c) 2002 Oliver Schwartz <Oliver.Schwartz@gmx.de>
 *                    Marko Friedemann <mfr@bmx-chemnitz.de>,
 *
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * This driver is based on:
 *  - the original xpad driver from Marko Friedmann
 *  - information from     http://euc.jp/periphs/xbox-controller.ja.html
 *  - the iForce driver    drivers/char/joystick/iforce.c
 *  - the skeleton-driver  drivers/usb/usb-skeleton.c
 *
 * Thanks to:
 *  - ITO Takayuki for providing essential xpad information on his website
 *  - Vojtech Pavlik     - iforce driver / input subsystem
 *  - Greg Kroah-Hartman - usb-skeleton driver
 *
 *
 * History:
 *
 * 2002-08-31 - 0.0.1 : first version, based on v0.0.6 of xpad driver
 * 2002-09-03 - 0.0.2 : Use a periodic timer for movement reporting
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/usb.h>
#include <linux/version.h>
#include <linux/timer.h>

#define DRIVER_VERSION "v0.0.2"
#define DRIVER_AUTHOR "Oliver Schwartz <Oliver.Schwartz@gmx.de>, Marko Friedemann <mfr@bmx-chemnitz.de>"
#define DRIVER_DESC "X-Box pad mouse driver"

#define XPAD_PKT_LEN 32

static struct xpad_device {
	u16 idVendor;
	u16 idProduct;
	char *name;
} xpad_device[] = {
	{ 0x045e, 0x0202, "Microsoft X-Box pad (US)" },
	{ 0x045e, 0x0285, "Microsoft X-Box pad (Japan)" },
	{ 0x05fd, 0x107a, "InterAct 'PowerPad Pro' X-Box pad (Germany)" },
	{ 0x0000, 0x0000, "X-Box pad" }
};

static const short dyn_table[] = {-6, -4, -3, -2, -2, -1, -1, 0,
                                   0,  1,  1,  2,  2,  3,  4, 6};

static struct usb_device_id xpad_table [] = {
	{ USB_INTERFACE_INFO('X', 'B', 0) },    /* X-Box USB-IF not approved class */
	{ }
};

MODULE_DEVICE_TABLE (usb, xpad_table);

struct usb_xpad {
	struct input_dev dev;			/* input device interface */
	struct usb_device *udev;		/* usb device */

	struct urb *irq_in;			/* urb for interrupt in report */
	unsigned char idata[XPAD_PKT_LEN];	/* input data */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	char phys[65];				/* physical device path */
#endif
	int open_count;				/* reference count */
	struct timer_list timer;
	signed short rel_wheel, rel_x, rel_y;
	unsigned char btn_left, btn_right, btn_middle;
};


static void xpad_timer(unsigned long data)
{
	struct usb_xpad * xpad = (struct usb_xpad *)data;
	/* cheap trick: report every event twice to make mouse more reactive */
	input_report_rel(&xpad->dev, REL_WHEEL, xpad->rel_wheel);
	input_report_rel(&xpad->dev, REL_X, xpad->rel_x);
	input_report_rel(&xpad->dev, REL_Y, xpad->rel_y);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 29)
	input_sync(dev);
#endif
	add_timer(&xpad->timer);
}

/*
 *	xpad_process_packet
 *
 *	Completes a request by converting the data into events for the
 *	input subsystem.
 *
 *	The used report descriptor was taken from ITO Takayukis website:
 *	 http://euc.jp/periphs/xbox-controller.ja.html
 */
static void xpad_process_packet(struct usb_xpad *xpad, u16 cmd, unsigned char *data)
{
	struct input_dev *dev = &xpad->dev;

	/* left stick */
	xpad->rel_x = dyn_table[(__s8)data[13]/16+8];
	xpad->rel_y = -dyn_table[(__s8)data[15]/16+8];

	/* right stick */
//	input_report_abs(dev, ABS_RX, (__s16) (((__s16)data[17] << 8) | data[16]));
	xpad->rel_wheel = dyn_table[(__s8)data[19]/32];

	// buttons
	input_report_key(dev, BTN_LEFT, data[4]);
	input_report_key(dev, BTN_RIGHT, data[5]);
	input_report_key(dev, BTN_MIDDLE, data[6]);

}

static void xpad_irq_in(struct urb *urb)
{
	struct usb_xpad *xpad = urb->context;

	if (urb->status)
		return;
	xpad_process_packet(xpad, 0, xpad->idata);
}

static int xpad_open (struct input_dev *dev)
{
	struct usb_xpad *xpad = dev->private;

	if (xpad->open_count++)
		return 0;

	init_timer(&xpad->timer);
	xpad->timer.expires = 1*HZ/5; /* every 200 ms */
	xpad->timer.data = (unsigned long)xpad;
	xpad->timer.function = xpad_timer;
	add_timer(&xpad->timer);
	xpad->irq_in->dev = xpad->udev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	if (usb_submit_urb(xpad->irq_in, GFP_KERNEL))
#else
	if (usb_submit_urb(xpad->irq_in))
#endif
		return -EIO;

	return 0;
}

static void xpad_close (struct input_dev *dev)
{
	struct usb_xpad *xpad = dev->private;

	if (!--xpad->open_count)
	{
		del_timer(&xpad->timer);
		usb_unlink_urb(xpad->irq_in);
	}
}

static void * xpad_probe(struct usb_device *udev, unsigned int ifnum, const struct usb_device_id *id)
{
	struct usb_xpad *xpad = NULL;
	struct usb_endpoint_descriptor *ep_irq_in;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	char path[64];
#endif
	int i;

	for (i = 0; xpad_device[i].idVendor; i++) {
		if ((udev->descriptor.idVendor == xpad_device[i].idVendor) &&
		    (udev->descriptor.idProduct == xpad_device[i].idProduct))
			break;
	}

	if ((xpad = kmalloc (sizeof(struct usb_xpad), GFP_KERNEL)) == NULL) {
		err("cannot allocate memory for new pad");
		return NULL;
	}
	memset(xpad, 0, sizeof(struct usb_xpad));

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	xpad->irq_in = usb_alloc_urb(0, GFP_KERNEL);
#else
	xpad->irq_in = usb_alloc_urb(0);
#endif
        if (!xpad->irq_in) {
		err("cannot allocate memory for new pad irq urb");
                kfree(xpad);
                return NULL;
        }

	ep_irq_in = udev->actconfig->interface[ifnum].altsetting[0].endpoint + 0;

	FILL_INT_URB(xpad->irq_in, udev,
		usb_rcvintpipe(udev, ep_irq_in->bEndpointAddress),
		xpad->idata, XPAD_PKT_LEN, xpad_irq_in, xpad,
		ep_irq_in->bInterval);

	xpad->udev = udev;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 29)
	xpad->dev.id.bustype = BUS_USB;
	xpad->dev.id.vendor = udev->descriptor.idVendor;
	xpad->dev.id.product = udev->descriptor.idProduct;
	xpad->dev.id.version = udev->descriptor.bcdDevice;
#else
	xpad->dev.idbus = BUS_USB;
	xpad->dev.idvendor = udev->descriptor.idVendor;
	xpad->dev.idproduct = udev->descriptor.idProduct;
	xpad->dev.idversion = udev->descriptor.bcdDevice;
#endif
	xpad->dev.private = xpad;
	xpad->dev.name = xpad_device[i].name;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	xpad->dev.phys = xpad->phys;
#endif
	xpad->dev.open = xpad_open;
	xpad->dev.close = xpad_close;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	usb_make_path(udev, path, 64);
	snprintf(xpad->phys, 64,  "%s/input0", path);
#endif

	xpad->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_REL);
	xpad->dev.evbit[0] = BIT(EV_KEY) | BIT(EV_REL);
	xpad->dev.keybit[LONG(BTN_MOUSE)] = BIT(BTN_LEFT) | BIT(BTN_RIGHT) | BIT(BTN_MIDDLE);
	xpad->dev.relbit[0] = BIT(REL_X) | BIT(REL_Y);
	xpad->dev.keybit[LONG(BTN_MOUSE)] |= BIT(BTN_SIDE) | BIT(BTN_EXTRA);
	xpad->dev.relbit[0] |= BIT(REL_WHEEL);


	input_register_device(&xpad->dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
	printk(KERN_INFO "input: %s on %s", xpad->dev.name, path);
#else
	printk(KERN_INFO "input: %s", xpad->dev.name);
#endif

	return xpad;
}

static void xpad_disconnect(struct usb_device *udev, void *ptr)
{
	struct usb_xpad *xpad = ptr;

	usb_unlink_urb(xpad->irq_in);
	input_unregister_device(&xpad->dev);
	usb_free_urb(xpad->irq_in);
	kfree(xpad);
}

static struct usb_driver xpad_driver = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 29)
	.name		= "xpad-mouse",
	.probe		= xpad_probe,
	.disconnect	= xpad_disconnect,
	.id_table	= xpad_table,
#else
	name:		"xpad-mouse",
	probe:		xpad_probe,
	disconnect:	xpad_disconnect,
	id_table:	xpad_table,
#endif
};

static int __init usb_xpad_init(void)
{
	usb_register(&xpad_driver);
	info(DRIVER_DESC ":" DRIVER_VERSION);
	return 0;
}

static void __exit usb_xpad_exit(void)
{
	usb_deregister(&xpad_driver);
}

module_init(usb_xpad_init);
module_exit(usb_xpad_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
