#include <linux/kernel.h>
#include <linux/sched.h>	//needed for 'event'
#include <linux/blkdev.h>	//needed for 'blk_size'
#include <linux/locks.h>
#include <linux/fatx_fs.h>
#include <linux/smp_lock.h>

#define PRINTK(format, args...) do { if (fatx_debug) printk( format, ##args ); } while(0)

//#define PRINTK(format, args...) printk( format, ##args )

//static inline void do_nothing() {}
//#define PRINTK(format, args...) do_nothing()

/* doesn't deal with root inode */
static void fatx_fill_inode(struct inode *inode, struct fatx_dir_entry *de)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int nr;

	INIT_LIST_HEAD(&MSDOS_I(inode)->i_fat_hash);
	MSDOS_I(inode)->i_location = 0;
	MSDOS_I(inode)->i_fat_inode = inode;
	inode->i_uid = sbi->options.fs_uid;
	inode->i_gid = sbi->options.fs_gid;
	inode->i_version = ++event;
	inode->i_generation = CURRENT_TIME;
	
	if ((de->attr & ATTR_DIR) && !FATX_IS_FREE(de)) {
		inode->i_generation &= ~1;
		inode->i_mode = MSDOS_MKMODE(de->attr,S_IRWXUGO &
		    ~sbi->options.fs_umask) | S_IFDIR;
		inode->i_op = sbi->dir_ops;
		inode->i_fop = &fatx_dir_operations;

		MSDOS_I(inode)->i_start = CF_LE_L(de->start);
		MSDOS_I(inode)->i_logstart = MSDOS_I(inode)->i_start;
		inode->i_nlink = fatx_subdirs(inode) + 2;
		    /* includes .., compensating for "self" */
#ifdef DEBUG
		if (!inode->i_nlink) {
			printk("directory %d: i_nlink == 0\n",inode->i_ino);
			inode->i_nlink = 1;
		}
#endif
		if ((nr = MSDOS_I(inode)->i_start) != 0)
			while (nr != -1) {
				inode->i_size += 1 << sbi->cluster_bits;
				if (!(nr = fat_access(sb, nr, -1))) {
					printk("Directory %ld: bad FAT\n",
					    inode->i_ino);
					break;
				}
			}
		MSDOS_I(inode)->mmu_private = inode->i_size;
	} else { /* not a directory */
		inode->i_generation |= 1;
		inode->i_mode = MSDOS_MKMODE(de->attr,
		    ((sbi->options.showexec)	//mark all as exec. if asked
		    	? S_IRUGO|S_IWUGO : S_IRWXUGO)
		    & ~sbi->options.fs_umask) | S_IFREG;
		MSDOS_I(inode)->i_start = CF_LE_L(de->start);
		MSDOS_I(inode)->i_logstart = MSDOS_I(inode)->i_start;
		inode->i_size = CF_LE_L(de->size);
	        inode->i_op = &fat_file_inode_operations;
	        inode->i_fop = &fat_file_operations;
		inode->i_mapping->a_ops = &fat_aops;
		MSDOS_I(inode)->mmu_private = inode->i_size;
	}
	if(de->attr & ATTR_SYS)
		if (sbi->options.sys_immutable)
			inode->i_flags |= S_IMMUTABLE;
	MSDOS_I(inode)->i_attrs = de->attr & ATTR_UNUSED;
	/* this is as close to the truth as we can get ... */
	inode->i_blksize = 1 << sbi->cluster_bits;
	inode->i_blocks = ((inode->i_size + inode->i_blksize - 1)
			   & ~(inode->i_blksize - 1)) >> 9;
	inode->i_mtime = inode->i_atime =
		fatx_date_dos2unix(CF_LE_W(de->time),CF_LE_W(de->date));
	inode->i_ctime =
		MSDOS_SB(sb)->options.isvfat
		? fatx_date_dos2unix(CF_LE_W(de->ctime),CF_LE_W(de->cdate))
		: inode->i_mtime;
//	MSDOS_I(inode)->i_ctime_ms = de->ctime_ms;	//FIXME: for fatx
}

struct inode *fatx_build_inode(
		struct super_block *sb,
		struct fatx_dir_entry *de,
		int ino,
		int *res )
{
	struct inode *inode;
	*res = 0;
	inode = fat_iget(sb, ino);
	if (inode)
		goto out;
	inode = new_inode(sb);
	*res = -ENOMEM;
	if (!inode)
		goto out;
	*res = 0;
	inode->i_ino = iunique(sb, MSDOS_ROOT_INO);
	fatx_fill_inode(inode, de);
	fat_attach(inode, ino);
	insert_inode_hash(inode);
out:
	return inode;
}

static unsigned int fatx_get_total_size ( struct super_block *sb )
{
	int major = MAJOR(sb->s_dev);
	
	if (blk_size[major]) {
		int minor = MINOR(sb->s_dev);
		return blk_size[major][minor] << 1;
	}
	return 0;
}

/*
 * parse super block values out of FATX "boot block"
 * unlike the other FAT variants, much of the data is calculated from the
 * the partition information.
 */
int fatx_parse_boot_block ( 
		struct super_block *sb, 
		struct buffer_head *bh )
{
	struct fatx_boot_sector *b = (struct fatx_boot_sector *)bh->b_data;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int logical_sector_size, hard_blksize;
	unsigned int rootdir_sectors, total_sectors;
	unsigned long fat_length;
	int spc_bits;

	PRINTK("FATX: entered fatx_parse_boot_block\n");
	
	if (b->magic != FATX_BOOTBLOCK_MAGIC) {
		printk("FATX: boot block signature not found.  Not FATX?\n");
		return -1;
	}
		
	PRINTK("FATX: fatx_magic: %08lX\n",(unsigned long)b->magic);
			
	logical_sector_size = 512;
	
	sbi->cluster_size = b->cluster_size;
	if (!sbi->cluster_size
	    || (sbi->cluster_size & (sbi->cluster_size - 1))) {
		printk("FATX: bogus cluster size %d\n", sbi->cluster_size);
		return -1;
	}
	
	PRINTK("FATX: cluster_size: %d\n",(int)b->cluster_size);
	
	//sb->s_block_size enters as hardware block (sector) size
	hard_blksize = sb->s_blocksize;
	sb->s_blocksize = logical_sector_size;
	sb->s_blocksize_bits = ffs(logical_sector_size) - 1;

	//I don't think this can happen with FATX, but I'll leave it in case.	
	if (logical_sector_size < hard_blksize) {
		printk("FATX: logical sector size too small for device"
		       " (logical sector size = %d)\n", logical_sector_size);
		return -1;
	}
	
	//figure total sector count
	total_sectors = fatx_get_total_size(sb);
	
	PRINTK("FATX: total_sectors for given device: %ld\n",(unsigned long)total_sectors);
	
	total_sectors -= 8;	//remove our boot block sectors from our calcs
	
	sbi->cluster_bits = ffs(logical_sector_size * sbi->cluster_size) - 1;
	sbi->fats = b->fats;	//will always be 1 (from observations)
	
	//hmm...fat should start right after boot block sectors (first 8)
	sbi->fat_start = 8;	//this might be: + CF_LE_W(b->fatx_unknown)
	sbi->root_cluster = 0;
	sbi->dir_per_block = logical_sector_size/sizeof(struct fatx_dir_entry);
	sbi->dir_per_block_bits = ffs(sbi->dir_per_block) - 1;
	sbi->dir_entries = 256;
	
	rootdir_sectors = sbi->dir_entries
		* sizeof(struct fatx_dir_entry) / logical_sector_size;
	
	//ok, I'm hoping this is right...it seems logical...
	//largest possible number of FAT table sectors in FATX-16 is:
	//   (65536 entries)*(2 bytes/entry)/(512 bytes/sector) == 256 sectors
	//check to see if it *could* use fatx-16...if so it does
	
	//regular fat16/32 has a fudge factor...the # of clusters beyond those
	//that can be addressed to ignore before switching over to fat32
	//TODO: determine if MSDOS_MAX_EXTRA (or similar) is used in fatX
	
	//bit shift to take sector count to cluster count (will be 5)
	spc_bits = sbi->cluster_bits - sb->s_blocksize_bits;
	PRINTK("FATX: spc_bits: %d (should be 5)\n",(int)spc_bits);
	
	//check cluster count
	if ((total_sectors-256)>>spc_bits <= 65536UL) {
		//FATX-16
		sbi->fat_bits = 16;	//I'm assuming there's no FATX-12
	} else {
		//FATX-32
		sbi->fat_bits = 32;
		
		//FUCKING HACK!  no clue why fat32 is handled differently
		//probably to allow > 1 cluster in root dir.
		sbi->root_cluster = 1;
	}

	sbi->fat_length = ((((total_sectors>>spc_bits)		//fat entries
	                * (sbi->fat_bits>>3)) 			//used bytes
	                + 4095) & ~4095)			//alloced bytes
	                / logical_sector_size;			//sectors
	
	PRINTK("FATX: fat_length(1): %d\n",(int)sbi->fat_length);
	
	//ok, the real magic
	//it's possible, with the subtraction of the space taken
	//up by the fat table, that a smaller table would work.  To
	//test this we recalculate subtracting the next smaller fatx
	//table size and see if the resulting number of clusters is 
	//representable in that smaller fat table.
		
	//next smallest fat table length (always in 4k blocks)		
	fat_length = sbi->fat_length - 8;
		
	fat_length = (((((total_sectors-fat_length)>>spc_bits)
	           * (sbi->fat_bits>>3))
	           + 4095) & ~4095)
	           / logical_sector_size;
	
	PRINTK("FATX: fat_length(2): %d\n",(int)fat_length);
	
	if (fat_length == sbi->fat_length - 8) {
		//this was a borderline case.  adjust fat_length
		sbi->fat_length = fat_length;
	}
		
	sbi->dir_start = sbi->fat_start + sbi->fats * sbi->fat_length;
	sbi->data_start = sbi->dir_start + rootdir_sectors;
	sbi->clusters = ((total_sectors+8) - sbi->data_start) / sbi->cluster_size;
	sbi->free_clusters = -1; /* Don't know yet */
	
	PRINTK("FATX: fat_start:  %d\n",(int)sbi->fat_start);
	PRINTK("FATX: dir_start:  %d\n",(int)sbi->dir_start);
	PRINTK("FATX: data_start: %d\n",(int)sbi->data_start);
	PRINTK("FATX: clusters:   %ld\n",(unsigned long)sbi->clusters);
	PRINTK("FATX: fat_bits:   %d\n",(int)sbi->fat_bits);
	PRINTK("FATX: fat_length: %d\n",(int)sbi->fat_length);
	PRINTK("FATX: root_dir_sectors: %d\n",(int)rootdir_sectors);
	PRINTK("FATX: dir_per_block:    %d\n",(int)sbi->dir_per_block);
	PRINTK("FATX: cluster_bits:     %d\n",(int)sbi->cluster_bits);
	
	if ( !sbi->fats 
		|| (sbi->dir_entries & (sbi->dir_per_block - 1))
//		|| sbi->clusters + 2 > fat_clusters + MSDOS_MAX_EXTRA //needed?
		|| logical_sector_size < 512
		|| PAGE_CACHE_SIZE < logical_sector_size ) {
		return -1;
	}

	PRINTK("FATX: leaving fatx_parse_boot_block\n");
		
	sbi->options.isfatx = 1;
	
	return 0;
}

static void fatx_read_root(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int nr;

	INIT_LIST_HEAD(&MSDOS_I(inode)->i_fat_hash);
	MSDOS_I(inode)->i_location = 0;
	MSDOS_I(inode)->i_fat_inode = inode;
	inode->i_uid = sbi->options.fs_uid;
	inode->i_gid = sbi->options.fs_gid;
	inode->i_version = ++event;
	inode->i_generation = 0;
	inode->i_mode = (S_IRWXUGO & ~sbi->options.fs_umask) | S_IFDIR;
	inode->i_op = sbi->dir_ops;
	inode->i_fop = &fatx_dir_operations;
	if (sbi->fat_bits == 32) {
		MSDOS_I(inode)->i_start = sbi->root_cluster;
		if ((nr = MSDOS_I(inode)->i_start) != 0) {
			while (nr != -1) {
				inode->i_size += 1 << sbi->cluster_bits;
				if (!(nr = fat_access(sb, nr, -1))) {
					printk("Directory %ld: bad FAT\n",
					       inode->i_ino);
					break;
				}
			}
		}
	} else {
		MSDOS_I(inode)->i_start = 0;
		inode->i_size = sbi->dir_entries * sizeof(struct fatx_dir_entry);
	}
	inode->i_blksize = 1 << sbi->cluster_bits;
	inode->i_blocks = ((inode->i_size + inode->i_blksize - 1)
			   & ~(inode->i_blksize - 1)) >> 9;
	MSDOS_I(inode)->i_logstart = 0;
	MSDOS_I(inode)->mmu_private = inode->i_size;

	MSDOS_I(inode)->i_attrs = 0;
	inode->i_mtime = inode->i_atime = inode->i_ctime = 0;
	MSDOS_I(inode)->i_ctime_ms = 0;
	inode->i_nlink = fatx_subdirs(inode) + 2;
}

/* The public inode operations for the fatx fs */
struct inode_operations fatx_dir_inode_operations = {
	create:		fatx_create,
	lookup:		fatx_lookup,
	unlink:		fatx_unlink,
	mkdir:		fatx_mkdir,
	rmdir:		fatx_rmdir,
	rename:		fatx_rename,
	setattr:	fat_notify_change,
};

void fatx_write_inode(struct inode *inode, int wait)
{
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct fatx_dir_entry *raw_entry;
	unsigned int i_pos;
	
	PRINTK("FATX: fatx_write_inode: entered\n");

retry:
	i_pos = MSDOS_I(inode)->i_location;
	if (inode->i_ino == MSDOS_ROOT_INO || !i_pos) {
		return;
	}
	lock_kernel();
	if (!(bh = fat_bread(sb, i_pos >> MSDOS_SB(sb)->dir_per_block_bits))) {
		PRINTK("dev = %s, ino = %d\n", kdevname(inode->i_dev), i_pos);
		fat_fs_panic(sb, "fatx_write_inode: unable to read i-node block");
		unlock_kernel();
		return;
	}
	spin_lock(&fat_inode_lock);
	if (i_pos != MSDOS_I(inode)->i_location) {
		spin_unlock(&fat_inode_lock);
		fat_brelse(sb, bh);
		unlock_kernel();
		goto retry;
	}

	raw_entry = &((struct fatx_dir_entry *) (bh->b_data))
	    [i_pos & (MSDOS_SB(sb)->dir_per_block - 1)];
	if (S_ISDIR(inode->i_mode)) {
		raw_entry->attr = ATTR_DIR;
		raw_entry->size = 0;
	}
	else {
		raw_entry->attr = ATTR_NONE;
		raw_entry->size = CT_LE_L(inode->i_size);
	}
	raw_entry->attr |= MSDOS_MKATTR(inode->i_mode) |
	    MSDOS_I(inode)->i_attrs;
	raw_entry->start = CT_LE_L(MSDOS_I(inode)->i_logstart);
	
	PRINTK("FATX: fatx_write_inode: start == %08lX (LE=%08lX)\n",
			(long)MSDOS_I(inode)->i_logstart,
			(long)CT_LE_L(MSDOS_I(inode)->i_logstart));
	
	fatx_date_unix2dos(inode->i_mtime,&raw_entry->time,&raw_entry->date);
	raw_entry->time = CT_LE_W(raw_entry->time);
	raw_entry->date = CT_LE_W(raw_entry->date);
	
	fatx_date_unix2dos(inode->i_ctime,&raw_entry->ctime,&raw_entry->cdate);
	raw_entry->ctime = CT_LE_W(raw_entry->ctime);
	raw_entry->cdate = CT_LE_W(raw_entry->cdate);
	
	spin_unlock(&fat_inode_lock);
	fat_mark_buffer_dirty(sb, bh);
	fat_brelse(sb, bh);
	unlock_kernel();
	
	PRINTK("FATX: fatx_write_inode: leaving\n");
}

int fatx_statfs(struct super_block *sb,struct statfs *buf)
{
	int retv = fat_statfs(sb,buf);
	
	buf->f_namelen = 42;
	return retv;
}

/*
static void fatx_delete_inode(struct inode *inode)
{
	printk("FATX: delete_inode entered, nlink = %d\n",(int)inode->i_nlink);
	fat_delete_inode(inode);
	printk("FATX: delete_inode leaving\n");
}

static void fatx_clear_inode(struct inode *inode)
{
	printk("FATX: clear_inode entered\n");
	fat_clear_inode(inode);
	printk("FATX: clear_inode leaving\n");
}

static void fatx_put_super(struct super_block *sb)
{
	printk("FATX: put_super entered\n");
	fat_put_super(sb);
	printk("FATX: put_super leaving\n");
}

static struct super_operations fatx_sops = { 
	write_inode:	fatx_write_inode,
	delete_inode:	fatx_delete_inode,
	put_super:	fatx_put_super,
	statfs:		fatx_statfs,
	clear_inode:	fatx_clear_inode,

	read_inode:	make_bad_inode,
};
/*/
static struct super_operations fatx_sops = { 
	write_inode:	fatx_write_inode,
	delete_inode:	fat_delete_inode,
	put_super:	fat_put_super,
	statfs:		fatx_statfs,
	clear_inode:	fat_clear_inode,

	read_inode:	make_bad_inode,
};
//*/		
		
struct super_block *fatx_read_super(struct super_block *sb,void *data, int silent)
{
	struct super_block *res;

	PRINTK("FATX: entering fatx_read_super\n");
	
	MSDOS_SB(sb)->options.isvfat = 0;
	res = fat_read_super_boot_block(
			sb, data, silent, 
			&fatx_dir_inode_operations,
			fatx_parse_boot_block,
			fatx_read_root);
	if (res) {
		sb->s_root->d_op = &fatx_dentry_operations;
		sb->s_op = &fatx_sops;
	}

	PRINTK("FATX: leaving fatx_read_super\n");
	
	return res;
}
